
from ..forward import Sigmoid,ReLU,Tanh,Linear,Identity 
import numpy as np

class MultiClassPerceptron:

    def __init__(self , y_out , y_true ):

        self.y_true= y_true-1
        self.y_out=y_out

        self.grads={}


    def loss(self):

        sample_loss=0
        
        last_layer=len(self.layers_num_arr)
        a_prelast= self._params['A'+ str(last_layer-1)] 
        z=self._params['Z'+ str(last_layer)]
        dW = np.zeros((len(z), len(a_prelast)+1))
        dA_prev = np.zeros((len( z), len(a_prelast)))
        a_prelast= np.append( a_prelast ,1).T
        w=self._params['W'+ str(last_layer)] 
        act_fc=self.layer_activations[last_layer]
        dl_dz=(np.zeros(len(z))).reshape(len(z) , 1)

        label=self.y_true

        if act_fc == Sign or act_fc == Identity or act_fc == None:

            m = max(z)
            max_indexes=[i for i, j in enumerate(z) if j == m]

            if label not in max_indexes:
                for i in max_indexes:
                    dW[i] = a_prelast 
                    dA_prev[i] = w[i]
                    dl_dz[i] = 1
        
                dW[label] = -a_prelast
                dA_prev[label] = w[label]
                dl_dz[label] = -1
                sample_loss = m - z[label]

            dl_db=dW[:,-1]
            dl_db=dl_db.reshape(len( z),1)

        
            self.grads['dW'+str(last_layer)] = dW[:,:-1]
            self.grads['db'+str(last_layer)] =dl_db
            self.grads['dA'+str(last_layer)] = (np.dot(dA_prev.T , dl_dz)).reshape(len(a_prelast)-1 ,1)


       
    

        else:
            m = max( self.y_out)
            max_indexes=[i for i, j in enumerate( self.y_out) if j == m]

            if label not in max_indexes:
                for i in max_indexes:
                    dl_dz[i] =act_fc.backwards( self.y_out[i]) 
                    dW[i] =  a_prelast * dl_dz[i]
                    dA_prev[i] = w[i]
                
                dl_dz[label] = -act_fc.backwards( self.y_out[label]) 

                dW[label] =  a_prelast * dl_dz[label]
                dA_prev[label] = w[label]     
                
                sample_loss = m -  self.y_out[label]
                
            dl_db=dW[:,-1]
            dl_db=dl_db.reshape(len( z),1)

        
            self.grads['dW'+str(last_layer)] = dW[:,:-1]
            self.grads['db'+str(last_layer)] =dl_db
            self.grads['dA'+str(last_layer)] = (np.dot(dA_prev.T , dl_dz)).reshape(len(a_prelast)-1 ,1)



        
        return sample_loss 





import numpy as np
from itertools import count

class Layer_Dense:
    _ids = count(1)
    _params= dict()#{}
    layer_activations= dict()#{}
    layers_num_arr = list()#[]

    def __init__(self, n_inputs, n_neurons, weight_type="random"):
        '''
        # Initialize weights and biases randomly
        @params :   n_inputs ---> number of features
                    n_neurons --> number of neurons
        '''


        self.layer_number = next(Layer_Dense._ids)
        Layer_Dense.layers_num_arr.append(self.layer_number)

        
        # Initializing the weights either random or zeros
        if weight_type == "random":
            self.weights = 0.1 * np.random.randn( n_neurons,n_inputs)
        elif weight_type == "zeros":
           self.weights = np.zeros(( n_neurons, n_inputs))

        self.biases = np.zeros((n_neurons, 1))

        # adding the initialized weight to _params dictionary
        Layer_Dense._params['W'+ str(self.layer_number)] = self.weights
        Layer_Dense._params['b'+ str(self.layer_number)] = self.biases
    def __del__(self):
        Layer_Dense._params.clear()
        Layer_Dense.layer_activations.clear()
        Layer_Dense.layers_num_arr.clear() #= 5#.clear()
        Layer_Dense._ids = count(1)


    #   passing data from activation function classes to Layer_Dens class
    def pass_act(self,act_type: str , A):
        """
        description 
        @Param: ---
        @Return: ---
        """
        Layer_Dense._params['A'+ str(self.layer_number)] = A
        Layer_Dense.layer_activations[self.layer_number] = act_type

    # multibly W*X and save them to out and _params dictionary
    def forward(self,inputs):
        if self.layer_number == 1: 
            Layer_Dense._params['A0'] = inputs
        self.out = np.dot(self.weights, inputs) + self.biases # Z

        return (self.out, self.weights,self.biases) 
        
    # get weights
    def get_weights(self):
        return self.weights

    # set weights and save them to _params dictionary
    def set_weights(self , new_weights):
        self.weights = new_weights
        Layer_Dense._params['W'+ str(self.layer_number)] = self.weights
        Layer_Dense._params['b'+ str(self.layer_number)] = self.biases

    def __reper__(self):
        pass



import math
import numpy as np 

class Adam:
    def __inti__(self):
        pass
    @staticmethod
    def initialize_adam(parameters,layerlen):
        L = layerlen  # number of layers in the neural networks
        v = {}
        s = {}
        
        for l in range(L):
            v["dW" + str(l+1)] = np.zeros(parameters['W'+str(l+1)].shape)
            v["db" + str(l+1)] = np.zeros(parameters['b'+str(l+1)].shape)
            s["dW" + str(l+1)] = np.zeros(parameters['W'+str(l+1)].shape)
            s["db" + str(l+1)] = np.zeros(parameters['b'+str(l+1)].shape)
        
        return v, s
    @staticmethod   
    def update_parameters(layerlen,parameters, grads, v, s, t=2, learning_rate = 0.01,
                                beta1 = 0.9, beta2 = 0.999,  epsilon = 1e-8):
    
        L = layerlen                 # number of layers in the neural networks
        v_corrected = {}             # Initializing first moment estimate, python dictionary
        s_corrected = {}              # Initializing second moment estimate, python dictionary
        
        # Perform Adam update on all parameters
        for l in range(L):
            # Moving average of the gradients. Inputs: "v, grads, beta1". Output: "v".
            v["dW" + str(l+1)] = beta1*v["dW" + str(l+1)] + (1-beta1)*grads['dW' + str(l+1)]
            v["db" + str(l+1)] = beta1*v["db" + str(l+1)] + (1-beta1)*grads['db' + str(l+1)]

            # Compute bias-corrected first moment estimate. Inputs: "v, beta1, t". Output: "v_corrected".
            v_corrected["dW" + str(l+1)] = v["dW" + str(l+1)]/(1-np.power(beta1,t))
            v_corrected["db" + str(l+1)] = v["db" + str(l+1)]/(1-np.power(beta1,t))

            # Moving average of the squared gradients. Inputs: "s, grads, beta2". Output: "s".
            s["dW" + str(l+1)] = beta2*s["dW" + str(l+1)]+(1-beta2)*np.power(grads['dW' + str(l+1)],2)
            s["db" + str(l+1)] = beta2*s["db" + str(l+1)]+(1-beta2)*np.power(grads['db' + str(l+1)],2)

            # Compute bias-corrected second raw moment estimate. Inputs: "s, beta2, t". Output: "s_corrected".
            s_corrected["dW" + str(l+1)] = s["dW" + str(l+1)]/(1-np.power(beta2,t))
            s_corrected["db" + str(l+1)] = s["db" + str(l+1)]/(1-np.power(beta2,t))
            
            # Update parameters. Inputs: "parameters, learning_rate, v_corrected, s_corrected, epsilon". Output: "parameters".
            parameters["W" + str(l+1)] = parameters["W" + str(l+1)] - (learning_rate*v_corrected["dW" + str(l+1)]/
                                                            np.sqrt(s_corrected["dW" + str(l+1)]+epsilon))
            parameters["b" + str(l+1)] = parameters["b" + str(l+1)] - (learning_rate*v_corrected["db" + str(l+1)]/
                                                            np.sqrt(s_corrected["db" + str(l+1)]+epsilon))        

        return parameters


import os
import shutil
import tarfile
import urllib.request
import pandas as pd

CIFAR10_URL = 'https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz'


class CIFAR_10:
    def __init__(self, path, download=True, train=True):
        self.path = path
        self.download = download
        self.train = train
        self.csv_list = []
        if self.download:
            self._Download()
        self.path = os.getcwd() + '/' + self.path
        self.toCSV()

        self.TrainFile = self.path + '/' + 'cifar-10-batches-py/train_cifar.csv'
        self.TestFile = self.path + '/' + 'cifar-10-batches-py/test_batch.csv'

    def _Download(self):
        if not os.path.exists(os.getcwd() + '/' + self.path):
            os.mkdir(self.path)
        file_name = 'CIFAR-10.tar.gz'
        with urllib.request.urlopen(CIFAR10_URL) as response, open(os.getcwd() + '/' + self.path + '/' + file_name,
                                                                   'wb') as out_file:
            shutil.copyfileobj(response, out_file)
            tar = tarfile.open(os.getcwd() + '/' + self.path + '/' + file_name, "r:gz")
            tar.extractall(os.getcwd() + '/' + self.path + '/')
            tar.close()

    def unpickle(self, file):
        import pickle
        with open(file, 'rb') as fo:
            dict = pickle.load(fo, encoding='bytes')
        return dict

    def toCSV(self):
        file_names = ['data_batch_1', 'data_batch_2', 'data_batch_3', 'data_batch_4', 'data_batch_5', 'test_batch']
        for name in file_names:
            df_labels = pd.DataFrame(self.unpickle(self.path + '/' + 'cifar-10-batches-py/' + name)[b'labels'])
            df_data = pd.DataFrame(self.unpickle(self.path + '/' + 'cifar-10-batches-py/' + name)[b'data'])
            new = pd.concat([df_labels, df_data], axis=1)
            if not os.path.exists(self.path + '/' + 'cifar-10-batches-py/' + name + '.csv'):
                new.to_csv(self.path + '/' + 'cifar-10-batches-py/' + name + '.csv', index=False)

        for name in file_names[0:5]:
            self.csv_list.append(self.path + '/' + 'cifar-10-batches-py/' + name + '.csv')


        if not os.path.exists(self.path + '/' + 'cifar-10-batches-py/train_cifar.csv'):
            df_from_each_file = (pd.read_csv(f, sep=',', header=None) for f in self.csv_list)
            df_merged = pd.concat(df_from_each_file, ignore_index=True)
            df_merged.to_csv(self.path + '/' + 'cifar-10-batches-py/train_cifar.csv', index=False)

    def __repr__(self):
        return self.TrainFile if self.train == True else self.TestFile






import pandas
import numpy as np
import math

class DataLoaderIterator:
    def __init__(self,dataloader):
        self.dataloader = dataloader
        self.index = 0
    def __next__(self):
        if self.index < self.dataloader.iternums:
            result = self.dataloader.__next__() 
            self.index +=1
            return result[0],result[1] 
        self.dataloader.current_position = 0    
        raise StopIteration
        
class DataLoader:
    def __init__(self,path='',batchsize=20,shuffling=True, normalization={}):
        """
        initializing the params
        @params: path
                batchsize      ---> number of raws loaded from csv
                shuffling      ---> shuffle raws indexs
                nomaliztion    ---> normalize between 0-1

        @return: 1) features raws
                2) labels raws
        """
        self.path = str(path)
        self.current_position = 0
        self.batchsize = batchsize
        self.shuffling = shuffling
        self.normalization = normalization['Transform'] if type(normalization['Transform']) == bool else RuntimeError('should be bool value ') 
        fileObject = open(self.path)
        row_count = sum(1 for row in fileObject) 
        self.iternums = math.ceil((row_count-1) / batchsize) - 5

    def __iter__(self):
           return DataLoaderIterator(self) 
    def __next__(self):
            df = pandas.read_csv(self.path, skiprows=self.current_position,nrows=self.batchsize)  
            self.current_position += self.batchsize

            if self.normalization == False:             #if No normalization

                if not self.shuffling:                  #if No shuffling
                    return df.iloc[:, 1:].to_numpy(), df.iloc[:, 0].to_numpy().reshape(-1,1)

                else:                                   #if shuffling
                    x = df.iloc[:, 1:].to_numpy()
                    y = df.iloc[:, 0].to_numpy()
                    np.random.shuffle(x)
                    np.random.shuffle(y)
                    return x , y.reshape(-1,1)
                
            else:                                       #if normalization
                if not self.shuffling:                  #if No shuffling
                    x = df.iloc[:, 1:].to_numpy()
                    norm_x = (x - np.min(x))/np.ptp(x)
                    return norm_x , df.iloc[:, 0].to_numpy().reshape(-1,1)

                else:                                   #if shuffling
                    x = df.iloc[:, 1:].to_numpy()
                    norm_x = (x - np.min(x))/np.ptp(x)
                    y = df.iloc[:, 0].to_numpy()
                    np.random.shuffle(norm_x)
                    np.random.shuffle(y)
                    return norm_x , y.reshape(-1,1)






import numpy as np


class SD():
    def __init__(self,pred,lable,learning_rate=0.1):
        self.pred = pred
        self.lable = lable
        self.parameters = 1 #Layer_Dense._params #Layer_Dense._params 
        self.dl={}
        
    def loss(self) :    
      pred_minus_lable = np.subtract(self.pred , self.lable)
      pred_minus_lable_T = pred_minus_lable.T
      return 0.5 * np.dot(pred_minus_lable_T, pred_minus_lable)

    def grad(self) :
      return np.subtract(self.pred , self.lable)


    @staticmethod
    def _Loss(prediction,label) :  
      lossval = 0.5*(prediction-label)**2
      return lossval
    
    @staticmethod
    def _Cost(prediction,label):
      m = label.shape[1] # datasize 
      cost = (1/m) * np.sum(SD._Loss(prediction,label))
      cost = np.squeeze(cost)
      return cost 

    @staticmethod
    def _Grad(prediction,label) :
      return np.subtract(prediction,label)  

    




import numpy as np

class Momentum:
    def __init__(self,paramterts,lr=0.1,beta=0.9):
        self.params = paramterts
        self.lr = lr
        self.beta = beta 
        self.v = {}
    @staticmethod 
    def initialize_velocity(parameter,layerlen):
        L = layerlen 
        v = {}
        for l in range(L):
            v['dW'+str(l+1)] = np.zeros(parameter['W'+str(l+1)].shape) 
            v['db'+str(l+1)] = np.zeros(parameter['b'+str(l+1)].shape)
        return v
    @staticmethod    
    def update(layerlen,v,parameter,lr,beta=0.9):
        L = layerlen  
        for l in range(L):
            v["dW" + str(l+1)] = (beta*v["dW" + str(l+1)]) + ((1-beta)*parameter['W'+str(l+1)])
            v["db" + str(l+1)] = (beta*v["db" + str(l+1)]) + ((1-beta)*parameter['b'+str(l+1)])
            parameter["W" + str(l+1)] = parameter["W" + str(l+1)] - (lr*v["dW" + str(l+1)])
            parameter["b" + str(l+1)] = parameter["b" + str(l+1)] - (lr*v["db" + str(l+1)])
        return parameter
    def __repr__(self):
        self.initialize_velocity()
        self.update()
        return {'velocity':self.v,
                    'parameters':self.params}    

                



from .BiploarPerceptron import BiploarPerceptron 
from .SD import SD 
from .Biploar_SVM import Biploar_SVM 
from .SoftmaxCrossEntropy import SoftmaxCrossEntropy
from .MultiClassPerceptron import  MultiClassPerceptron


import numpy as np

class Linear:
    
    def __init__(self, n_inputs, n_neurons,l_size, weight_type="random_l"):
        '''
        # Initialize weights and biases randomly
        @params :   n_inputs ---> number of features
                    n_neurons --> number of neurons
        '''
        # Initializing the weights either random or zeros
        if weight_type == "random":
            self.weights = 0.1 * np.random.randn(n_neurons,n_inputs)
        elif weight_type == "random_l":
            self.weights = np.random.randn(n_neurons,n_inputs) * np.sqrt(2/l_size)     
        elif weight_type == "zeros":
           self.weights = np.zeros(( n_neurons, n_inputs))

        self.biases = np.zeros((n_neurons, 1))
   
    def forward(self,X):
        Z = np.dot(self.weights,X) + self.biases 
        return Z 

    def set_weights(self , new_weights):
        self.weights = new_weights

    def set_b(self,b):
        self.biases = b    
    def updateW_B(self,weights,bias):
        self.weights = weights 
        self.biases = bias
    def __reper__(self):
        return self.weights, self.biases
    def Values(self):
        return self.weights, self.biases



import os
import zipfile, urllib.request, shutil
import requests


MNIST_URL = 'https://drive.google.com/uc?id=1NjvEw9Ob7sJkEQLWPe_M-XhhZLYCx7lE&export=download'


class FashionMNIST:
    def __init__(self,path,download=True,train=True):
        self.path = path
        self.download = download
        self.train = train
        if self.download:
            self._Download()
         
        self.TrainFile = os.getcwd() + '/' + self.path + '/mnist_train.csv'
        self.TestFile = os.getcwd() + '/' + self.path + '/mnist_test.csv'

    def _Download(self):
        if not os.path.exists(os.getcwd() + '/' + self.path):
            os.mkdir(self.path)
        file_name = 'MNIST.zip'
        with urllib.request.urlopen(MNIST_URL) as response, open(os.getcwd() + '/' + self.path + '/' + file_name,
                                                                 'wb') as out_file:
            shutil.copyfileobj(response, out_file)
            with zipfile.ZipFile(os.getcwd() + '/' + self.path + '/' + file_name) as zf:
                zf.extractall(os.getcwd() + '/' + self.path + '/')

    def __repr__(self):
        return self.TrainFile if self.train == True else self.TestFile


from .Linear import Linear
from . import Sigmoid
from . import ReLU,Sigmoid,Tanh,Softmax,Identity,Softmax
from itertools import repeat
from ..losses import SD#, Biploar_SVM ,BiploarPerceptron
from ..backward import Backward
import numpy as np
from ..losses import SoftmaxCrossEntropy
from ..optimizations import Optimizer


def _tupleDivide(tup):
    layersShape =  list()    
    for i in range(len(tup)-1):
        _tup = (tup[i],tup[i+1])
        layersShape.append(_tup)
    return layersShape

class NetWork:
    def __init__(self,LayersShape,activations,optimType={'GD':True}):
        self.LayersShape = LayersShape
        self.createdLayers = [Linear(layer[0],layer[1],l_size=layer[0]) for layer in _tupleDivide(LayersShape)]
        self.activations = list(repeat(activations, len(self.createdLayers))) if type(activations) == str \
                                                else activations 

        self.LayersLen = len(self.createdLayers) 
        self.parameters = dict()
        self.Zout = list()
        self.Aout = list()
        self.lossvalues = []
        self.optimType  = optimType 

    def fit(self,X,Y,learning_rate=0.1):
        ########## Forward L Step        
        self._CalcFWD(X)
        self._ConstructParams() 

        ############## Lossess
        # cost = SD._Loss(self.Aout[self.LayersLen-1],Y)
        # cost = SD._Cost(self.Aout[self.LayersLen-1],Y)
        cost = SoftmaxCrossEntropy._Cost(self.Aout[self.LayersLen-1],Y)
        
        # cost = compute_cost(self.Aout[self.LayersLen-1],Y)
        # loss = lossClass._Loss(self.Aout[self.LayersLen-1],Y)


        # lossD = SD._Grad(self.Aout[self.LayersLen-1],Y)
        # lossD = lossgrad(self.Aout[self.LayersLen-1],Y)
        lossD = SoftmaxCrossEntropy._Grad(self.Aout[self.LayersLen-1],Y.T)
        # Y label 
        # lossD = lossClass._Grad(self.Aout[self.LayersLen-1],Y)
        
        
        ########## backward 
        grads = Backward.StaticBW(self.parameters,learning_rate=learning_rate,LossDerivative=lossD,
                    Y=Y,layersLen=self.LayersLen,layer_activations=self.activations)
        ###########
        
        ######### parameters update
        optimizer = Optimizer(parameters=self.parameters,grads=grads,optimiType=self.optimType,LayersLen=self.LayersLen,lr=learning_rate)
        self.parameters = optimizer.optimize()
        self._UpdateLayerParam()
        ###########################

        self.Aout.clear()
        self.Zout.clear()
        return cost  
    

    def Parameters(self):
        return self.parameters    
    def _updateA_Z(self):
        for i in range(len(self.createdLayers)):
            self.Zout[i] = self.parameters['Z'+str(i+1)]
            self.Aout[i] = self.parameters['A'+str(i+1)]


    def _UpdateLayerParam(self):
        for i in range(self.LayersLen):
            self.createdLayers[i].updateW_B(self.parameters['W'+str(i+1)],self.parameters['b'+str(i+1)])

    def _ConstructParams(self):
        for i in range(self.LayersLen):
            self.parameters['W'+str(i+1)],self.parameters['b'+str(i+1)] = self.createdLayers[i].__reper__()
            self.parameters['Z'+str(i+1)] = self.Zout[i]
            self.parameters['A'+str(i+1)] = self.Aout[i]

    def _CalcFWD(self,X):
        aOut = X
        self.parameters['A0'] = X
        for Layer,activation in zip(self.createdLayers,self.activations):
            zOut = Layer.forward(aOut)
            self.Zout.append(zOut)
            aOut = NetWork.ActivationCalc(zOut,activation)
            self.Aout.append(aOut) 
    
    def FWD_predict(self,X):
        aOut = X
        for Layer,activation in zip(self.createdLayers,self.activations):
            zOut = Layer.forward(aOut)
            aOut = NetWork.ActivationCalc(zOut,activation)
        return aOut  

    def Prediction(self,X,Y,parameter):
        if  self.LayersShape[-1] > 1:
            return self._MultiPrediction(X,Y,parameter)
        return self._BinaryPrediction(X,Y,parameter)
    # multi classfication       
    def _MultiPrediction(self,X,Y,parameter):
        m = X.shape[1]
        p = list()#np.zeros((1,m))
        probas = self.FWD_predict(X)
        for i in range(0, probas.shape[1]):
            m = max(probas[:,i])
            max_indexes=[s for s, t in enumerate(probas[:,i]) if t == m]
            p.append(max_indexes[0])
    
        return p
    # binay classfication    
    def _BinaryPrediction(self,X,Y,parameter):
        m = X.shape[1]
        n = self.LayersLen # number of layers in the neural network
        p = np.zeros((1,m))
        
        # Forward propagation
        probas = self.FWD_predict(X)
        for i in range(0, probas.shape[1]):
            if probas[0,i] > 0.5:
                p[0,i] = 1
            else:
                p[0,i] = 0
        
        return p
        # pass 
    @staticmethod
    def ActivationCalc(zOut,activation='ReLU'):
        if activation == 'ReLU':
            return ReLU.ReLU_(zOut)
        elif activation == 'Sigmoid':
            return Sigmoid.sigmoid_(zOut)
        elif activation == 'Tanh':
            return Tanh.Tanh_(zOut) 
        elif activation == 'SoftMax':
            return Softmax.Softmax_(zOut)    

        assert 'No Activation Setted\n'         

    def PlotLoss(self):
        pass   



    def Linear(self,n_inputs, n_neurons, weight_type="random"):
        self.McreatedLayers.append(Linear(n_inputs, n_neurons, weight_type="random"))
        return self.McreatedLayers[0]
        pass
            
    def save_model(self,path):
        pass



import numpy as np
from .Layers import Layer_Dense

class Sigmoid:
    def forwards(self, inputs):
        X = inputs.out
        sig =   1 / (1 + (np.exp(-1 * X)))
        inputs.pass_act('Sigmoid',sig)
        return sig
    @staticmethod
    def sigmoid_(inputs):
        return 1 / (1 + (np.exp(-inputs)))
    @staticmethod    
    def sigmoidBW_(inputs):
        return Sigmoid.sigmoid_(inputs) * (1-Sigmoid.sigmoid_(inputs))    

    def backwards(self, inputs):
        s = self.forwards(inputs)
        return s * (1 - s)

class ReLU:

    def forwards(self, inputs):
        rel = np.maximum(0, inputs.out)
        inputs.pass_act('relu',rel)
        return rel
    @staticmethod     
    def ReLU_(inputs):
        return np.maximum(0, inputs)    
    @staticmethod
    def ReLUBW_(inputs):
        dZ = np.where(inputs > 0,1,inputs)
        dZ = np.where(dZ <= 0 ,0,dZ)
        return dZ
    def Backwards(self, inputs):
         if inputs > 0:
             return 1
         elif inputs <= 0:
             return 0

class Identity:
    def forwards(self, inputs):
        ident =inputs.out
        inputs.pass_act('identity',ident)
        return ident

    def backwards(self, inputs):
        pass

class Tanh:
    def forwards(self, inputs):
        tan = np.tanh(inputs.out)
        inputs.pass_act('tanh',tan)
        return tan
    @staticmethod
    def Tanh_(inputs):
        return np.tanh(inputs)
    @staticmethod 
    def TanhBW_(inputs):
        return (1 - Tanh.Tanh_(inputs)**2)

    def backwards(self, inputs):
        a = self.forwards(inputs)
        return 1 - a ** 2


class Softmax:
    @staticmethod
    def Softmax_(inputs):
        f = np.exp(inputs - np.max(inputs))  # shift values
        return f / f.sum()

    @staticmethod   
    def SoftmaxBW_(prediction,label):
        dy_dz=np.zeros_like(prediction)

        for i in range (label.shape[1]):
            ind=label[0,i]

            dy_dz[:,i]=prediction[:, i]*-prediction[ind, i]

            dy_dz[ind,i]+= prediction[ind, i]


        return   dy_dz 



import numpy as np 
from ..forward import Layer_Dense#,_params,layers_num_arr,layer_activations
from ..forward import ReLU,Sigmoid,Tanh,Softmax,Identity
from ..losses import SoftmaxCrossEntropy
from ..optimizations import Momentum,Adam


class Backward:  # make inheratnce to loss, and layers 
    def __init__(self,Y,learning_rate,LossDerivative):
        # self.AL = AL
        self.Y = Y
        self.grads = {} #grads 
        self.parameters = dict(Layer_Dense._params) #Layer_Dense._params #parameters 
        self.layers_num_arr  = list(Layer_Dense.layers_num_arr)
        self.layer_activations = dict(Layer_Dense.layer_activations)
        self.learning_rate = learning_rate
        self.LossDerivative = LossDerivative
        self.caches = { i: (self.parameters['A'+str(i)],self.parameters['W'+str(i)],self.parameters['b'+str(i)]) if i != 0 else 
                                (self.parameters['A'+str(i)],'_') for i in range(0,int(((len(self.parameters.keys())-1)/3)+1))}
    def _L_model_backward(self):# caches):

        L = len(self.layers_num_arr) # the number of layers
        m = self.caches[L][0].shape[1] 

        self.Y = self.Y.reshape(self.caches[L][0].shape)
        # Initializing the backpropagation
        dAL = self.LossDerivative # loss dervatie (dl/dy)
        self.grads['dA'+str(L)] = dAL
        current_cache = (self.caches[L-1][0],self.caches[L][1],self.caches[L][2]) 
        self.grads["dA" + str(L-1)], self.grads["dW" + str(L)], self.grads["db" + str(L)] = self._LinearActivaionBW(dAL,
                                    L,linear_cache=current_cache,activation=self.layer_activations[L])
        for l in reversed(range(1,L)):

            current_cache = (self.caches[l-1][0],self.caches[l][1],self.caches[l][2])#self.caches[l]

            dA_prev_temp, dW_temp, db_temp =  self._LinearActivaionBW(self.grads['dA'+str(l)],
                                    l,linear_cache=current_cache,activation=self.layer_activations[l])
            self.grads["dA" + str(l-1)] = dA_prev_temp
            self.grads["dW" + str(l)] = dW_temp
            self.grads["db" + str(l)] = db_temp

    def backward(self): # user call it 
        self._L_model_backward()
        return self._update_parameters(), self.grads
    

    def _LinearActivaionBW(self,dAL,l,linear_cache,activation='ReLU'):
        dZ = self._bacwardActivations(dAL,l=l,activation=activation) # (1,1)dz2
        dA_prev, dW, db = self._linear_backward(dZ,linear_cache)# da1,dw2,db2
        return dA_prev,dW,db

    def _bacwardActivations(self,dAL,l,activation='Relu'):        
        if activation == 'Relu':
            self.grads['dZ'+str(l)] = ReLU.ReLUBW_(dAL)  
        elif activation == 'Sigmoid':
            self.grads['dZ'+str(l)] = Sigmoid.sigmoidBW_(dAL)
            return self.grads['dZ'+str(l)]
        elif activation == 'Tanh':
            self.grads['dZ'+str(l)] = Tanh.TanhBW_(dAL)

    def _linear_backward(self,dZ, cache):        
        A_prev, W, b = cache
        m = A_prev.shape[1]
        dW = (1/m) * np.matmul(dZ,A_prev.T) # dz2 (1,1) * a1 (3*1)
        db = (1/m) * np.sum(dZ,axis=1,keepdims=True)

        dA_prev = np.matmul(W.T,dZ) # da1 = w2.t(3,1) . dz2(1,1) 
        assert (dA_prev.shape == A_prev.shape)
        assert (dW.shape == W.shape)
        assert (db.shape == b.shape)        
        return dA_prev, dW, db

    def _update_parameters(self):
               
        L = len(self.layers_num_arr) // 2 # number of layers in the neural network
        # Update rule for each parameter. Use a for loop.
        for l in range(L):
            self.parameters["W" + str(l+1)] =  self.parameters["W" + str(l+1)] - self.learning_rate * self.grads['dW'+str(l+1)]
            self.parameters["b" + str(l+1)] = self.parameters["b" + str(l+1)] - self.learning_rate * self.grads['db'+str(l+1)]
        return self.parameters  


    ################################################################
    ####################     Static Methods     ####################
    ################################################################
    @staticmethod
    def _L_Backward(layers_num_arr,parameters,Y,LossDerivative,layer_activations):# caches):   

        caches = { i: (parameters['A'+str(i)],parameters['W'+str(i)],parameters['Z'+str(i)],parameters['b'+str(i)]) if i != 0 else 
                                (parameters['A'+str(i)],'_') for i in range(0,layers_num_arr+1)}     
        L = layers_num_arr # the number of layers 
        m = caches[L][0].shape[1]  
        grads = dict()
        # Initializing the backpropagation
        if layer_activations[L-1]  != 'SoftMax':
            dAL = LossDerivative  
            grads['dA'+str(L)] = dAL 
            current_cache = (caches[L-1][0],caches[L][1],caches[L][2]) 
            grads["dA" + str(L-1)], grads["dW" + str(L)],grads["db" + str(L)] = Backward._StaticLinearActivaionBW(dAL,
                                            L,parameter=parameters,Y=Y,linear_cache=current_cache,activation=layer_activations[L-1],grads=grads)
        else: 
            dZ = SoftmaxCrossEntropy._Grad(caches[L][0],Y)#.sum(axis=1,keepdims=True)      # dl/dy * dy/dz
            current_cache = (caches[L-1][0],caches[L][1],caches[L][2])
            grads["dA" + str(L-1)], grads["dW" + str(L)],grads["db" + str(L)] = Backward._StaticLinearBW(dZ,current_cache)


        for l in reversed(range(L-1)):
            current_cache = (caches[l][0],caches[l+1][1],caches[l+1][2])

            dA_prev_temp, dW_temp, db_temp = Backward._StaticLinearActivaionBW(grads['dA'+str(l+1)],
                                    l+1,Y=Y,linear_cache=current_cache,parameter=parameters,activation=layer_activations[l],grads=grads)
            grads["dA" + str(l)] = dA_prev_temp
            grads["dW" + str(l+1)] = dW_temp
            grads["db" + str(l+1)] = db_temp

        return grads  
    @staticmethod
    def StaticBW(parameters,learning_rate,LossDerivative,Y,layersLen,layer_activations): # user call it 
        grads = Backward._L_Backward(layers_num_arr=layersLen,parameters=parameters,Y=Y,LossDerivative=LossDerivative,layer_activations=layer_activations)
        return grads
        
    @staticmethod
    def _StaticLinearActivaionBW(dAL,l,linear_cache,parameter,Y,activation='',grads=''):
        dZ = Backward._StaticBWActivations(parameter=parameter,Y=Y,l=l,activation=activation,grads=grads) # (1,1)dz2
        dZL  = Backward._multiply(dAL,dZ) # multiply 
        dA_prev, dW, db = Backward._StaticLinearBW(dZL,linear_cache)# da1,dw2,db2
        return dA_prev,dW,db
    @staticmethod
    def _StaticBWActivations(l,parameter,activation='ReLU',grads='',Y=0):  
        if activation == 'ReLU':
            grads['dZ'+str(l)] = ReLU.ReLUBW_(parameter['Z'+str(l)]) 
            return grads['dZ'+str(l)]
        elif activation == 'Sigmoid':
            grads['dZ'+str(l)] = Sigmoid.sigmoidBW_(parameter['Z'+str(l)]) 
            return grads['dZ'+str(l)]
        elif activation == 'Tanh':
            grads['dZ'+str(l)] = Tanh.TanhBW_(parameter['Z'+str(l)])
            return grads['dZ'+str(l)]
        elif activation == 'SoftMax':
            grads['dZ'+str(l)] = Softmax.SoftmaxBW_(parameter['A'+str(l)],Y)
            return grads['dZ'+str(l)]
    @staticmethod 
    def _multiply(A,Z):
        return A * Z
    @staticmethod
    def _StaticLinearBW(dZ,cache):
        A_prev, W, b = cache
        m = A_prev.shape[1]
        dW =  np.matmul(dZ,A_prev.T) 
        db =  np.sum(dZ,axis=1,keepdims=True)
        dA_prev = np.matmul(W.T,dZ)  
        assert (dA_prev.shape == A_prev.shape)
        assert (dW.shape == W.shape)        
        return dA_prev, dW, db



import numpy as np


class MultiPerceptron:

    @staticmethod
    def _Cost(prediction,label):

        lossval=0

        for i in range (label.shape[1]):

            ind=label[0,i]

            m = max(prediction[:,i])
            max_indexes=[s for s, t in enumerate(prediction[:,i]) if t == m]

            ind= label[0,i]
            
            if ind not in max_indexes:
                

                lossval+=prediction[max_indexes[0]][i]-prediction[ind][i]

        cost = lossval/label.shape[1]

        return cost
                
       

    
    @staticmethod
    def _Grad(prediction,label) :
        
        
        dl_dy=np.zeros_like(prediction)

        for i in range (label.shape[1]):

            m = max(prediction[:,i])
            max_indexes=[s for s, t in enumerate(prediction[:,i]) if t == m]

            ind= label[0,i]
            
            if ind not in max_indexes:
                
                dl_dy[ind][i]=-1

                for j in max_indexes:
                    dl_dy[j][i]=1
        
        
        return dl_dy

    
    
        
    



# import numpy as np

# class multi_class_SVM:


#     @staticmethod
#     def _Cost(prediction,label):

#         lossval=0

#         for i in range(label.shape[1]):

#             ind=label[0,i]

#             y_label=prediction[ind][i]
            
#             for j in range(prediction.shape[0]):

#                 if max(1+prediction[j][i]-y_label , 0) and j!= ind:

#                     lossval+=1+prediction[j][i]-y_label

#         m = label.shape[1] # datasize 
#         cost = lossval/m

#         return cost

        

          

    
    
#     @staticmethod
#     def _Grad(prediction,label) :
        
        
#         dl_dy=np.zeros_like(prediction)

#         for i in range (label.shape[1]):
#             ind=label[0,i]

#             y_label=prediction[ind][i]

#             count = 0
            
#             for j in range(prediction.shape[0]):

#                 if max(1+prediction[j][i]-y_label , 0) and j!= ind:

#                     count+=1
#                     dl_dy[j][i]=1
                    

#             if count>0:

#                 dl_dy[ind][i]=-count

    
        
#         return  dl_dy

    
    
   



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '0.1.2'
PACKAGE_NAME = 'TourchPIP'
AUTHOR = ''
AUTHOR_EMAIL = ''
URL = 'https://github.com/Mostafa-ashraf19/TourchPIP'

LICENSE = 'MIT'
DESCRIPTION = 'DL Framework help you to build simple neural network'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'numpy',
      'pandas',
      'matplotlib',
      'requests'
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages(),
      download_url='https://github.com/Mostafa-ashraf19/TourchPIP/archive/v_01.2.tar.gz'
      )



import numpy as np
from ..forward import Layer_Dense
from ..forward import Sigmoid,ReLU,Tanh,Identity


class Biploar_SVM(Layer_Dense) :
    def __init__(self,pred,lable):
        self.pred = pred
        self.lable = lable
        self.parameters = Layer_Dense._params #parameter
        self.dl={}
   
    def loss(self):
        return max(0,1-(self.lable*self.parameters['Z'+str(len(Layer_Dense.layer_activations.keys()))]))
        
    def grad(self):
        L = len(Layer_Dense.layer_activations.keys())
        act_fc=Layer_Dense.layer_activations[L]
   
        if (self.lable*self.parameters['Z'+str(L)] )> 1:
                self.dl['dW'+str(L)] = np.zeros(self.parameters['W'+str(L)].shape)
                self.dl['db'+str(L)] = np.zeros(self.parameters['b'+str(L)].shape)
                self.dl['dA'+str(L)] = np.zeros(self.parameters['A'+str(L)].shape)
        else :
            dl_activation=[]
            if act_fc == 'sigmoid':
               sig_inst=Sigmoid()
               dl_activation = sig_inst.backwards(self.parameters['Z'+str(L)]) 
  
            elif act_fc == 'relu':
                relu_inst =ReLU()
                dl_activation= relu_inst.Backwards(self.parameters['Z'+str(L)])
            elif act_fc == 'tanh':
                tanh_inst = Tanh()
                dl_activation= tanh_inst.backwards(self.parameters['Z'+str(L)])
            dl_dz = np.matmul(-self.lable,dl_activation)
            self.dl['dW'+str(L)] = np.matmul(self.parameters['W'+str(L)],dl_dz)
            self.dl['db'+str(L)] = dl_dz
           
    @staticmethod
    def _Loss(parameters,label) :  
      lossval = max(0,1-(label*parameters['Z'+str(len(Layer_Dense.layer_activations.keys()))]))
      return lossval
    
    @staticmethod
    def _Cost(prediction,label):
      m = label.shape[1] # datasize 
      cost = (1/m) * np.sum(Biploar_SVM._Loss(prediction,label))
      cost = np.squeeze(cost)
      return cost 

    @staticmethod
    def _Grad(parameters,label) :
        L = len(Layer_Dense.layer_activations.keys())
        dl_dy=[]
        if (label*parameters['Z'+str(L)] )> 1:   
          dl_dy = 0
               
        else :
            dl_dy=-label
           
        return dl_dy

            
            



from . import Momentum , Adam

class Optimizer:
    def __init__(self,parameters,optimiType={'GD':True},LayersLen=0,grads=[],lr=0):
        self.optimiType = optimiType
        self.grads = grads 
        self.lr = lr
        self.LayersLen = LayersLen
        self.parameters = parameters

    def normalGD(self):
        L = self.LayersLen  # number of layers in the neural network
        # Update rule for each parameter. Use a for loop.
        for l in range(L):
            self.parameters["W" + str(l+1)] = self.parameters["W" + str(l+1)] - self.lr * self.grads['dW'+str(l+1)]
            self.parameters["b" + str(l+1)] = self.parameters["b" + str(l+1)] - self.lr * self.grads['db'+str(l+1)]
        return self.parameters     
        
    def optimize(self):
            if 'GD' in self.optimiType and self.optimiType['GD']:
                return self.normalGD() # return new paramters 
            elif 'Adam' in self.optimiType and self.optimiType['Adam']:
                v,s = Adam.initialize_adam(parameters=self.parameters,layerlen=self.LayersLen)
                return Adam.update_parameters(layerlen=self.LayersLen,v=v,s=s,grads=self.grads,parameters=self.parameters,learning_rate=self.lr)
            elif 'Momeuntum' in self.optimiType and self.optimiType['Momeuntum']:
                v = Momentum.initialize_velocity(parameter=self.parameters,layerlen=self.LayersLen)
                return Momentum.update(layerlen=self.LayersLen,v=v,parameter=self.parameters,lr=self.lr)


import numpy as np

class SoftmaxCrossEntropy:
    def  __init__ (self):
        pass

    @staticmethod
    def _Cost(prediction,label):
        lossval=0
        for i in range (label.shape[1]):

            ind=label[0,i]
            y_label= prediction[ind][i]
            if y_label != 0:
                lossval+=-np.log(y_label)

        cost = lossval/label.shape[1]

        return cost
       

        
    @staticmethod
    def _Grad(prediction,label) :
        dl_dy=np.zeros_like(prediction)

        for i in range (label.shape[1]):
            ind=label[0,i]

            dl_dy[:,i]=prediction[:, i]

            dl_dy[ind,i]= prediction[ind, i]-1

        return   dl_dy

